from functools import partial

from qtpy.QtCore import Slot
from qtpy.QtWidgets import QMenu, QFileDialog

from fpga_device_manager import Devices, Pins, Popup, Inputs, Config, Constants
from fpga_device_manager.Devices import Output
from fpga_device_manager.Inputs import Input
from fpga_device_manager.device_base import DeviceBaseTemplate, DeviceBaseInstance
from fpga_device_manager.exceptions import InvalidConfigError
from fpga_device_manager.widgets.device_widget import OutputWidget, InputWidget
from fpga_device_manager.windows.base import BaseWindow
from fpga_device_manager.windows.device_settings import DeviceSettingsWindow


class MainWindow(BaseWindow):
    """The application's main window."""
    def __init__(self, *args, **kwargs):
        super(MainWindow, self).__init__("main.ui", *args, **kwargs)

        self.dirty = False
        self.device_widgets = {}

        self.init_btn_menu()
        self.init_colors()

        self.refresh()

    def set_dirty(self) -> None:
        """Sets the dirty flag, which prompts to save on several actions that would otherwise lose data."""
        self.dirty = True
        self.update_title()

    def clear_dirty(self) -> None:
        """Clears the dirty flag."""
        self.dirty = False
        self.update_title()

    def init_btn_menu(self) -> None:
        """Initializes the buttons for adding devices."""
        dev_menu = QMenu()
        input_menu = QMenu()

        for tpl_id, tpl in Devices.list_templates():
            dev_menu.addAction(f"Add {tpl.name}").triggered.connect(
                partial(self.add_device_template, tpl)
            )

        for tpl_id, tpl in Inputs.list_templates():
            input_menu.addAction("Add %s" % tpl.name).triggered.connect(
                partial(self.add_device_template, tpl, True)
            )

        self.btn_add_device.setMenu(dev_menu)
        self.btn_add_input.setMenu(input_menu)

        print(self.btn_add_device.menu())
        print(self.btn_add_input.menu())

    def init_colors(self) -> None:
        """Initializes the preview key colors."""
        for frame, color_type in {
            self.fr_reserved: Pins.PIN_RESERVED,
            self.fr_inputs: Pins.PIN_OCCUPIED_INPUT,
            self.fr_devices: Pins.PIN_OCCUPIED_DEVICE,
            self.fr_available: Pins.PIN_AVAILABLE,
            self.fr_available_pwm: Pins.PIN_AVAILABLE_PWM
        }.items():
            frame.setStyleSheet("background-color: rgb(%d, %d, %d);" % Pins.color(color_type)[0:3])

    def refresh(self) -> None:
        """Updates all widgets."""
        self.update_labels()
        self.update_buttons()
        self.update_preview()
        self.update_widgets()
        self.update_title()

    def update_labels(self) -> None:
        """Updates the window's labels."""
        self.tabs.setTabText(0, "Appliances (%d)" % (len(Config.outputs().devices)))
        self.tabs.setTabText(1, "Sensors (%d)" % (len(Config.inputs().devices)))

        self.lb_pins_used.setText(str(sum(pin.is_output() for pin in Pins.all())))
        self.lb_pins_input.setText(str(sum(pin.is_input() for pin in Pins.all())))
        self.lb_pins_free.setText(str(sum(not pin.is_assigned() and not pin.supports_pwm for pin in Pins.all())))
        self.lb_pins_free_pwm.setText(str(sum(not pin.is_assigned() and pin.supports_pwm for pin in Pins.all())))

    def update_buttons(self) -> None:
        """Updates the window's buttons."""
        self.btn_save.setEnabled(Config.outputs().has_devices() or Config.inputs().has_devices())
        self.btn_export.setEnabled(Config.outputs().has_devices() and Config.inputs().has_devices())
        self.btn_clear.setEnabled(Config.outputs().has_devices() or Config.inputs().has_devices())
        self.btn_add_device.setEnabled(not Config.outputs().has_max_devices())
        self.btn_add_input.setEnabled(not Config.inputs().has_max_devices())

    def update_widgets(self) -> None:
        """Updates all device widgets."""
        for widget in self.device_widgets.values():
            widget.refresh()

    def update_preview(self) -> None:
        """Updates the pin preview."""
        self.img_pins.update()

    def update_title(self) -> None:
        """Updates the window's title."""
        dirty = self.dirty and "*" or ""
        self.setWindowTitle(f"{Constants.APP_NAME} {Constants.APP_VERSION}{dirty}")

    def load(self, filename: str) -> None:
        """
        Loads a device configuration.
        :param filename: Name of configuration file to load
        """
        try:
            self.reset()
            Config.load(filename)

            self.add_all_widgets()
            self.clear_dirty()
            self.update_title()

        except Exception as e:
            Popup.alert(title=f"Failed to load {filename}",
                        message=str(e))
            raise

    def save(self, filename: str) -> None:
        """
        Saves the current device configuration.
        :param filename: Name of file to save to
        """
        try:
            Config.save(filename)
            self.clear_dirty()

        except Exception as e:
            Popup.error(title=f"Failed to save {filename}", ex=e)

    def export(self, path: str) -> None:
        """
        Exports the current device configuration to Verilog code.
        :param path: Path to output Verilog files to
        """
        try:
            Config.export(path)
            Popup.info(title="Successfully exported",
                       message=f"Verilog code has been successfully written to {path}.")

        except InvalidConfigError as e:
            Popup.alert(title=f"Failed to export",
                        message=str(e))

    def reset(self) -> None:
        """Resets the configuration and window."""
        Config.clear()
        self.remove_all_widgets()
        self.clear_dirty()

    def closeEvent(self, event) -> None:
        """
        Handler that gets executed when the window is about to close. If the dirty flag is set, this will prompt the
        user to save.
        :param event: Qt close event
        """
        if self.dirty:
            Popup.confirm(title="You have unsaved changes",
                          message="Are you sure you want to quit?",
                          additional_text="You will lose all unsaved changes.",
                          on_yes=lambda: event.accept(),
                          on_no=lambda: event.ignore())

    def add_device_template(self, tpl: DeviceBaseTemplate, is_input: bool = False) -> None:
        """
        Adds a new device, based on the supplied device template.
        :param tpl: The device template to base the new device off of
        :param is_input: Whether this is an input device or not
        """
        if is_input:
            dev_class = Input
            manager = Config.inputs()
            add_widget_call = self.add_input_widget
        else:
            dev_class = Output
            manager = Config.outputs()
            add_widget_call = self.add_output_widget

        try:
            device = dev_class.create_from_template(dev_name=tpl.name, dev_type=tpl.type)
            window = DeviceSettingsWindow(device)
            if window.exec_() == window.Accepted:
                for pin in device.pins.values():
                    try:
                        Pins.assign(Pins.get_next_available_pin(pin).name, pin)
                    except Exception as e:
                        print("could not assign pin %s: %s" % (pin.name, e))

                    if is_input:
                        try:
                            device.associate(next(Config.outputs().filter(device.template.compatible_devices)))
                        except StopIteration:
                            print("could not associate device: no compatible devices")
                manager.add_device(device)
                add_widget_call(device)
        except Exception as e:
            Popup.error(title=f"Failed to add device", ex=e)

    def add_output_widget(self, device: Output, refresh: bool = True) -> None:
        """
        Adds a widget for the specified output device.
        :param device: Output device to add widget for
        :param refresh: Whether the window should refresh after adding
        """
        widget = OutputWidget(device, main_window=self)
        self.device_widgets[device] = widget
        self.tab_devices.device_list.layout().addWidget(widget)
        if refresh:
            self.refresh()

    def add_input_widget(self, input_device: Input, refresh: bool = True) -> None:
        """
        Adds a widget for the specified input device.
        :param input_device: Input device to add widget for
        :param refresh: Whether the window should refresh after adding
        """
        widget = InputWidget(input_device, main_window=self)
        self.device_widgets[input_device] = widget
        self.tab_inputs.device_list.layout().addWidget(widget)
        if refresh:
            self.refresh()

    def add_all_widgets(self) -> None:
        """Adds widgets for all devices, based on the current configuration."""
        for mgr, call in ((Config.outputs(), self.add_output_widget), (Config.inputs(), self.add_input_widget)):
            for device in mgr.all_devices():
                call(device, refresh=False)

        self.refresh()

    def remove_widget(self, device: DeviceBaseInstance, refresh: bool = True) -> None:
        """
        Removes a widget belonging to a device. This is usually done to remove a device from the configuration.
        :param device: Device widget to remove
        :param refresh: Whether the window should refresh after removing
        """
        try:
            self.device_widgets[device].close()
            del self.device_widgets[device]
            if refresh:
                self.refresh()
        except KeyError:
            print(f"attempted to close nonexistent widget for device {device}")

    def remove_all_widgets(self) -> None:
        """Removes all device widgets."""
        for device in self.device_widgets.copy().keys():
            self.remove_widget(device, refresh=False)

        self.refresh()

    @Slot()
    def on_btn_clear_clicked(self):
        """Handler for clicking the Clear button. This prompts the user whether all devices should be removed."""
        Popup.confirm(title="Clear all devices",
                      message="Are you sure you wish to remove all devices and inputs?",
                      on_yes=self.reset)

    @Slot()
    def on_btn_load_clicked(self):
        """Handler for clicking the Load button. This opens a file selection dialog for loading a configuration."""
        filename = QFileDialog.getOpenFileName(parent=self,
                                               caption="Load device data",
                                               dir="./configurations",
                                               filter="JSON device data (*.json)")[0]

        # Cancelling results in filename being ''
        if filename != '':
            self.load(filename)

    @Slot()
    def on_btn_save_clicked(self):
        """Handler for clicking the Save button. This opens a file selection dialog for saving a configuration."""
        filename = QFileDialog.getSaveFileName(parent=self,
                                               caption="Load device data",
                                               dir="./configurations",
                                               filter="JSON device data (*.json)")[0]

        # Cancelling results in filename being ''
        if filename != '':
            self.save(filename)

    @Slot()
    def on_btn_export_clicked(self):
        """Handler for clicking the Export button. This performs some sanity checks on the current configuration,
        and if passed opens a directory selection dialog for exporting Verilog code.
        """
        # Verify device settings
        try:
            Config.check()
            path = QFileDialog.getExistingDirectory(parent=self,
                                                    caption="Export verilog code",
                                                    dir="./generated")

            if path != '':
                self.export(path)

        except InvalidConfigError as e:
            Popup.alert(title="Export failed",
                        message="\n".join(e.errors),
                        additional_text="Please check your configuration and try again.")
            return



"""Module for managing Output device templates. Also contains classes for instances of Output devices."""
from __future__ import annotations
import json
from typing import Iterable

from fpga_device_manager.device_base import DeviceBasePin, DeviceBaseInstance, DeviceBaseTemplate
from fpga_device_manager.device_params import DeviceParameter
from fpga_device_manager.exceptions import PinIncompatibilityError

_device_templates = {}


class DevicePin(DeviceBasePin):
    """Base class for device pins."""
    def __init__(self, requires_pwm=False, *args, **kwargs):
        super(DevicePin, self).__init__(*args, **kwargs)
        self.requires_pwm = requires_pwm

    def copy(self, device: DeviceBaseInstance) -> DevicePin:
        """
        Creates a copy of itself and assigns it to a device.
        :param device: Device the pin belongs to
        :return: Pin copy
        """
        return DevicePin(name=self.name,
                         requires_pwm=self.requires_pwm,
                         device=device,
                         active_low=self.active_low,
                         assigned_pin=self.assigned_pin)

    def check_validity(self) -> bool:
        """
        Verifies the configuration of this pin. Will throw a PinIncompatibilityError if the assigned pin is somehow
        incompatible with this pin.
        :return: True if configuration is valid
        """
        super(DevicePin, self).check_validity()

        if not self.is_compatible_with(self.assigned_pin):
            raise PinIncompatibilityError(self)

        return True

    def is_compatible_with(self, fpga_pin: 'FPGAPin') -> bool:
        """
        Returns whether this pin is compatible with the specified FPGA pin.
        :param fpga_pin: FPGA pin to check compatibility with
        :return: True if compatible, False otherwise
        """
        return fpga_pin.supports_output and not (self.requires_pwm and not fpga_pin.supports_pwm)

    def __str__(self):
        return "Pin %s of device %s" % (self.name, self.device.name)


class OutputTemplate(DeviceBaseTemplate):
    """Represents the blueprint on which Output devices can be based off of."""
    def __init__(self, uses_bus=False, *args, **kwargs):
        super(OutputTemplate, self).__init__(*args, **kwargs)
        self.uses_bus = uses_bus

    @staticmethod
    def load(dev_data: dict) -> OutputTemplate:
        """
        Creates an OutputTemplate based on loaded JSON data.
        :param dev_data: JSON dict template data
        :return: Template instance
        """
        obj = OutputTemplate(dev_name=dev_data["name"],
                             dev_type=dev_data["type"],
                             dev_driver=dev_data["driver"],
                             uses_clk=dev_data.get("uses_clk", True),
                             uses_bus=dev_data.get("uses_bus", False),
                             description=dev_data["description"],
                             parameters={
                                 param_name: DeviceParameter.load(param_name, param_data)
                                 for param_name, param_data in dev_data.get("parameters", {}).items()
                             })

        for pin_name, pin_data in dev_data["pins"].items():
            obj.pins[pin_name] = DevicePin(name=pin_name,
                                           device=obj,
                                           requires_pwm=pin_data["requires_pwm"],
                                           active_low=pin_data["active_low"])

        return obj


class Output(DeviceBaseInstance):
    """An output device or appliance."""
    def __init__(self, uses_bus=False, *args, **kwargs):
        super(Output, self).__init__(*args, **kwargs)
        self.associated_inputs = []
        self.uses_bus = uses_bus

    @staticmethod
    def create_from_template(dev_name: str, dev_type: int) -> Output:
        """
        Creates an Output device instance based on an OutputTemplate.
        :param dev_name: The name of the new device
        :param dev_type: The ID of the template to base the device off of
        :return: Output device instance
        """
        dev_template = get_template(dev_type)

        obj = Output(dev_name=dev_name,
                     dev_type=dev_template.type,
                     uses_clk=dev_template.uses_clk,
                     uses_bus=dev_template.uses_bus,
                     parameters={name: param.copy() for name, param in dev_template.parameters.items()})

        obj.template = dev_template

        for name, pin in dev_template.pins.items():
            obj.pins[name] = pin.copy(obj)

        return obj

    def save(self) -> dict:
        """
        Exports the device's state into a format that is JSON serializable.
        :return: JSON serializable dict
        """
        return {
            "name": self.name,
            "type": self.type,
            "pins": {
                pin.name: {
                    "assigned_pin": pin.assigned_pin.name if pin.assigned_pin is not None else None,
                    "active_low": pin.active_low
                } for pin in self.pins.values()
            },
            "parameters": {name: param.value for name, param in self.parameters.items()}
        }

    def remove_associations(self) -> None:
        """Removes all associated Input devices from this Output."""
        for input_dev in self.associated_inputs:
            input_dev.deassociate()

    def check_validity(self) -> bool:
        """
        Verifies the validity of this device's configuration by checking whether all of its pins are valid.
        :return: True if device is configured properly, False otherwise
        """
        for pin in self.pins.values():
            pin.check_validity()
        return True

    def __str__(self):
        return self.name


def init(filename: str) -> None:
    """
    Initializes the template module by loading and indexing all available Output device templates.
    :param filename: File to load template data from
    """
    global _device_templates
    try:
        with open(filename, "r") as file:
            data = json.load(file)
            for tpl_type, tpl_data in data.items():
                tpl_data["type"] = int(tpl_type)
                _device_templates[int(tpl_type)] = OutputTemplate.load(tpl_data)

    except OSError as e:
        raise Exception("Couldn't open device template file: %s" % e)


def list_templates() -> Iterable[OutputTemplate]:
    """
    Returns an Iterable of all Output device templates, indexed by ID.
    :return: Iterable of Output device templates
    """
    for tpl_id, tpl in _device_templates.items():
        yield tpl_id, tpl


def get_template(dev_type: int) -> OutputTemplate:
    """
    Looks up an Output device template by ID.
    :param dev_type: ID of Output device template
    :return: Output device template instance
    """
    try:
        return _device_templates[dev_type]
    except KeyError:
        raise Exception("No such device type: %s" % dev_type)



from fpga_device_manager.windows.base import BaseWidget


class DeviceListWidget(BaseWidget):
    """A widget that displays devices."""
    def __init__(self, *args, **kwargs):
        super(DeviceListWidget, self).__init__("device_list.ui", *args, **kwargs)



import argparse
import os
import sys

from fpga_device_manager import Devices, Inputs, Pins, Config
from fpga_device_manager.exceptions import InvalidConfigError, DeviceLoadError


def fail(err_msg: str = ""):
    if err_msg:
        sys.stderr.write(err_msg + "\n")
    sys.exit(1)


if __name__ == '__main__':
    argparser = argparse.ArgumentParser()

    argparser.add_argument("--output",
                           default="./generated",
                           help="Path where the generated files should be output to.")
    argparser.add_argument("config_file",
                           help="File containing device configuration.")

    args = vars(argparser.parse_args())

    base_data_path = os.path.join(os.path.dirname(__file__), "..", "res", "data")

    Devices.init(os.path.join(base_data_path, "device_types.json"))
    Inputs.init(os.path.join(base_data_path, "input_types.json"))
    Pins.init(os.path.join(base_data_path, "pins.json"))

    try:
        Config.load(args['config_file'])
    except DeviceLoadError as e:
        fail(f"An error occurred while loading the configuration: {e}")

    try:
        Config.check()
        Config.export(args['output'])

    except InvalidConfigError as e:
        error_str = ""
        for error in e.errors:
            error_str += f"\n - {error}"

        fail(f"The configuration is not valid: {error_str}")

    except Exception as e:
        fail(f"An error occurred: {e}")

    print(f"Successfully exported Verilog code to {os.path.abspath(args['output'])}")



"""This module holds all exceptions that can possibly occur during device management."""


class DeviceInvalidError(BaseException):
    """Gets thrown when a device, both Output or Input, has not been configured properly."""
    pass


class InvalidConfigError(BaseException):
    """
    Gets thrown when the device configuration as a whole did not pass vailidity checks.
    The errors parameter is a list of error messages that have been generated during verification.
    """
    def __init__(self, errors):
        super(InvalidConfigError, self).__init__()
        self.errors = errors


class PinNotAssignedError(DeviceInvalidError):
    """Gets thrown when a device pin has not been assigned to any FPGA pin."""
    def __init__(self, pin):
        super(PinNotAssignedError, self).__init__(f'Pin "{pin.name}" of device "{pin.device.name}" '
                                                  f'is not assigned to any FPGA pin')


class PinIncompatibilityError(DeviceInvalidError):
    """Gets thrown when a device pin has been assigned to an FPGA pin it is not compatible with."""
    def __init__(self, pin):
        super(PinIncompatibilityError, self).__init__(
            f'Pin "{pin.name}" of device "{pin.device.name}" '
            f'is not compatible with assigned pin "{pin.assigned_pin.name}"')


class DeviceIncompatibilityError(DeviceInvalidError):
    """Gets thrown when a pair of Input and Output devices are associated, but somehow incompatible with each other."""
    def __init__(self, dev_in, dev_out):
        super(DeviceIncompatibilityError, self).__init__(f'Devices "{dev_in.name}" and "{dev_out.name}" are not'
                                                         f'compatible with each other')


class InputNotAssociatedError(DeviceInvalidError):
    """Gets thrown when an Input device has not been assigned to any Output device."""
    def __init__(self, device):
        super(InputNotAssociatedError, self).__init__(f'Device "{device.name}" '
                                                      f'is not associated with any output device')


class DeviceLoadError(BaseException):
    """Gets thrown when an error occurs during loading of a device instance from a configuration."""
    pass


class UnknownPinException(DeviceLoadError):
    """Gets thrown during the attempt to load a device configuration that contains a device with a pin it shouldn't
    possess."""
    def __init__(self, device, pin):
        super(UnknownPinException, self).__init__(device,
                                                  f'Attempted to set up unknown pin "{pin.name}" '
                                                  f'on device "{device.name}.')



import re

from qtpy.QtCore import Slot
from qtpy.QtWidgets import QFormLayout

from fpga_device_manager import Popup
from fpga_device_manager.device_base import DeviceBaseInstance
from fpga_device_manager.windows.base import BaseDialog


class DeviceSettingsWindow(BaseDialog):
    """A dialog for setting a device's name and parameters."""
    def __init__(self, device: DeviceBaseInstance, *args, **kwargs):
        super(DeviceSettingsWindow, self).__init__("device_settings.ui", *args, **kwargs)
        self.device = device

        proto = device.template
        self.grp_properties.setTitle(proto.name)
        self.lb_description.setText(proto.description)
        self.txt_dev_name.setText(device.name or proto.name)

        self.param_widgets = {}
        self.setup_property_widgets()

    def setup_property_widgets(self):
        """Sets up widgets for all of the device's properties."""
        layout: QFormLayout = self.grp_properties.layout()
        for param_name, param in self.device.parameters.items():
            param_widget = param.create_widget()
            self.param_widgets[param_name] = param_widget
            layout.addRow(param.display_name, param_widget)

    def is_valid_name(self) -> bool:
        """
        Checks whether the chosen name is a valid name for the device.
        A name is valid if it contains between 1 and 60 alphanumeric characters, spaces or underscores.
        :return: True if the chosen name is valid
        """
        return re.search(r"^[a-zA-Z0-9_\- ]{1,60}$", self.txt_dev_name.text()) is not None

    @Slot()
    def accept(self):
        """Handler for pressing OK on the dialog.

        Validates the device's name and parameters.
        """
        # Validate name
        name = self.txt_dev_name.text()
        if not self.is_valid_name():
            Popup.alert(title="Invalid name",
                        message="The name \"%s\" is invalid." % name,
                        additional_text="A device name can only contain up 60 alphanumeric characters, hyphens, "
                                        "underscores and spaces and must not be empty.")
            return

        # Validate and set parameters
        try:
            for param_name, param in self.device.parameters.items():
                param.set_value(param.get_value_from_widget(self.param_widgets[param_name]))
        except Exception as e:
            Popup.alert(title="Invalid parameter", message=str(e))
            return

        # Set name and accept
        self.device.name = name
        super().accept()



"""A widget that draws the FPGA preview image, displaying pin headers and pin assignments."""
from collections import namedtuple
from math import ceil
from typing import Tuple

from qtpy.QtCore import QRect, Qt
from qtpy.QtGui import QPainter, QImage, QPixmap, QColor, QPen, QBrush, QFont
from qtpy.QtWidgets import QLabel

from fpga_device_manager import Pins

"""A pin header is defined as a tuple containing a name, pin group size and a list of tuples of pins."""
PinHeader = namedtuple("PinHeader", "name bank_width pins")

_PINS = [
    PinHeader("Pin header 1", 5, [
        ("34", "32", "GND", "27", "GND", "25", "23", "GND", "21", "19",
         "GND", "13", "11", "GND", "9", "5", "3", "1", "3.3", "3.3"),

        ("35", "33", "GND", "28", "GND", "26", "24", "GND", "22", "20",
         "GND", "14", "12", "GND", "10", "6", "4", "2", "NC", "IO3")
    ]),

    PinHeader("On-board LEDs", 4, [
        ("D8", "106", "105", "104", "100", "99", "98", "97")
    ])
]

_SIZE = (450, 200)
_BOX_BORDER_WIDTH = 4
_PIN_SIZE = 24
_PIN_BORDER_WIDTH = 3
_PIN_MARGIN = 6
_BANK_MARGIN = 16
_BOX_PADDING = 11


class PinHeaderImage(QImage):
    """The image of a single FPGA pin header."""
    def __init__(self, header: PinHeader):
        self.header = header
        width, height = self._header_box_size()

        super(PinHeaderImage, self).__init__(width, height * 2, QImage.Format_RGB32)

        self.painter = QPainter(self)
        self.pen = QPen(QColor(0, 0, 0), _BOX_BORDER_WIDTH, Qt.SolidLine)
        self.pen.setJoinStyle(Qt.MiterJoin)
        self.painter.setPen(self.pen)

        self._update()

    def _header_box_size(self) -> Tuple[int, int]:
        maxn = max(len(x) for x in self.header.pins)
        width = round(_PIN_SIZE * maxn +
                      (ceil(maxn / self.header.bank_width) - 1) * _BANK_MARGIN +
                      (maxn - ceil(maxn / self.header.bank_width)) * _PIN_MARGIN +
                      _BOX_PADDING * 2)
        height = (_PIN_SIZE * len(self.header.pins) +
                  _PIN_MARGIN * (len(self.header.pins) - 1) +
                  _BOX_PADDING * 2)
        return width, height

    def _update(self):
        self._clear()
        self._draw()

    def _clear(self):
        self.fill(QColor(255, 255, 255))

    def _draw(self):
        self._draw_bg()
        self._draw_pins()
        self.painter.end()

    def _draw_bg(self):
        self.pen.setWidth(_BOX_BORDER_WIDTH)
        self.painter.setPen(self.pen)
        self.painter.setBrush(QBrush(QColor(230, 230, 230)))

        border_offset = _BOX_BORDER_WIDTH // 2
        x = y = border_offset
        w, h = self._header_box_size()
        w -= border_offset * 2
        h -= border_offset * 2
        self.painter.drawRect(x, y, w, h)

    def _draw_pins(self):
        _, box_h = self._header_box_size()

        x = y = _BOX_PADDING

        font: QFont = self.painter.font()

        for pin_row in self.header.pins:
            for i, pin in enumerate(pin_row):
                # Determine pin color
                try:
                    pin_obj = Pins.get(pin)
                    pin_name = pin_obj.display_name
                    pin_color = pin_obj.color()
                    font_size = 18
                    font_bold = True

                except KeyError:
                    pin_name = pin
                    pin_color = Pins.color(Pins.PIN_RESERVED)
                    font_size = 11
                    font_bold = False

                font.setPixelSize(font_size)
                font.setBold(font_bold)
                self.painter.setFont(font)

                # Draw pin
                self.pen.setColor(QColor(0, 0, 0))
                self.pen.setWidth(_PIN_BORDER_WIDTH)
                self.painter.setPen(self.pen)
                self.painter.setBrush(QBrush(QColor(*pin_color)))
                border_offset = _PIN_BORDER_WIDTH // 2
                pin_x = x - border_offset
                pin_y = y - border_offset
                pin_w = pin_h = _PIN_SIZE - (_PIN_BORDER_WIDTH - border_offset)
                self.painter.drawRect(pin_x, pin_y, pin_w, pin_h)

                # Draw label + shadow
                self.pen.setColor(QColor(90, 90, 90))
                self.painter.setPen(self.pen)
                font_rect = QRect(x - 12, y + box_h - 12, _PIN_SIZE + 20, _PIN_SIZE + 20)
                self.painter.drawText(font_rect, Qt.AlignHCenter | Qt.AlignVCenter, pin_name)

                font_rect.setX(x - 14)
                font_rect.setY(y + box_h - 14)
                self.pen.setColor(QColor(*pin_color))
                self.painter.setPen(self.pen)
                self.painter.drawText(font_rect, Qt.AlignHCenter | Qt.AlignVCenter, pin_name)
                x += _PIN_SIZE + (_BANK_MARGIN if (i % self.header.bank_width == self.header.bank_width - 1)
                                  else _PIN_MARGIN)

            x = _BOX_PADDING
            y += _PIN_SIZE + _PIN_MARGIN


class FPGAPreview(QLabel):
    """The entire FPGA preview image."""
    def __init__(self, *args):
        super(FPGAPreview, self).__init__(*args)
        self.im_width = int(_SIZE[0] * 1.5)
        self.im_height = int(_SIZE[1] * 1.5)

        self.image = QImage(self.im_width, self.im_height, QImage.Format_RGB32)
        self.painter = QPainter(self.image)

        self.setFixedSize(*_SIZE)

        self.update()

    def update(self) -> None:
        """Updates the preview image with the current pin assignment data."""
        self._clear()
        self._draw()
        self.setPixmap(QPixmap(self.image))
        self.setScaledContents(True)

    def _clear(self):
        self.painter = QPainter(self.image)
        self.image.fill(QColor(255, 255, 255))

    def _draw(self):
        width = 0
        height = 0
        images = []

        # Draw all pin header images
        for header in _PINS:
            header_im = PinHeaderImage(header)
            images.append(header_im)
            width = max(width, header_im.width())
            height = height + header_im.height()

        y = (self.im_height - height) // 2

        # Copy pin header images onto own image
        for header_im in images:
            x = (self.im_width - header_im.width()) // 2

            self.painter.drawImage(x, y, header_im)
            y += header_im.height()

        self.painter.end()



from __future__ import annotations
from abc import ABC
from typing import Any, Tuple

from qtpy.QtWidgets import QWidget, QSpinBox, QSizePolicy
from qtpy.QtCore import Qt

from fpga_device_manager.widgets.colorbutton import ColorButton
from fpga_device_manager.widgets.percent_slider import PercentSlider
from fpga_device_manager.widgets.state_spinbox import StateSpinBox


class DeviceParameter(ABC):
    """Base class for device parameters."""
    def __init__(self, name, display_name=None, description="", value=None, default=None):
        self.name = name
        self.display_name = display_name or name
        self.description = description
        self.default = default
        self.value = value or default or 0

    @staticmethod
    def load(name: str, data: dict) -> DeviceParameter:
        """
        Factory for creating a new device parameter object from JSON data.
        Throws an Exception if the specified parameter type is unknown.
        :param name: Name of parameter
        :param data: JSON data
        :return: Parameter object
        """
        if data["type"] == "int":
            return IntParameter(name=name,
                                description=data.get("description", None),
                                display_name=data.get("name", None),
                                default=data.get("default", None),
                                v_min=data.get("min", 0),
                                v_max=data.get("max", 65535),)
        elif data["type"] == "percent":
            return PercentParameter(name=name,
                                    description=data.get("description", None),
                                    display_name=data.get("name", None),
                                    default=data.get("default", None))
        elif data["type"] == "state":
            return StateParameter(name=name,
                                  description=data.get("description", None),
                                  display_name=data.get("name", None),
                                  default=data.get("default", None))
        elif data["type"] == "color":
            return ColorParameter(name=name,
                                  description=data.get("description", None),
                                  display_name=data.get("name", None),
                                  default=data.get("default", None))
        raise Exception("Unknown parameter type %s" % data["type"])

    def copy(self) -> DeviceParameter:
        """
        Creates a copy of itself.
        :return: Copy parameter
        """
        raise NotImplementedError

    def create_widget(self) -> QWidget:
        """
        Creates a widget used for setting this parameter's value.
        :return: Parameter widget
        """
        widget = self._base_widget()
        widget.setToolTip(self.description)
        widget.sizePolicy().setHorizontalPolicy(QSizePolicy.Maximum)
        return widget

    def _base_widget(self) -> QWidget:
        """Returns the base widget for this parameter."""
        raise NotImplementedError

    def set_value(self, value: Any) -> None:
        """
        Assigns a new value to this parameter. If the value isn't found to be valid, this throws an Exception.
        :param value: The new value for this parameter
        """
        if not self.is_valid(value):
            raise Exception("%s is not a valid value for %s" % (value, self))

        self.value = value

    def get_value_from_widget(self, widget: QWidget) -> Any:
        """
        Extracts the current value from the parameter widget.
        :param widget: Widget to get value from
        """
        raise NotImplementedError

    def is_valid(self, value: Any) -> bool:
        """
        Verifies the validity of the given parameter value.
        :param value: Value to verify
        :return: True if value is valid, False otherwise
        """
        raise NotImplementedError

    def export(self) -> str:
        """Exports a string representation of the parameter's value."""
        raise NotImplementedError

    def __str__(self):
        return f"Parameter {self.name}"


class IntParameter(DeviceParameter):
    """An integer parameter."""
    def __init__(self, v_min, v_max, word_size=16, *args, **kwargs):
        super(IntParameter, self).__init__(*args, **kwargs)
        self.min = v_min
        self.max = v_max
        self.word_size = word_size

    def copy(self) -> IntParameter:
        """
        Creates a copy of itself.
        :return: Copy
        """
        return IntParameter(name=self.name,
                            display_name=self.display_name,
                            description=self.description,
                            v_min=self.min,
                            v_max=self.max,
                            default=self.default,
                            word_size=self.word_size,
                            value=self.value)

    def _base_widget(self) -> QSpinBox:
        """
        Returns a SpinBox with the parameter's settings.
        :return: SpinBox for this parameter
        """
        widget = QSpinBox()
        widget.setToolTip(self.description)
        widget.setMinimum(self.min)
        widget.setMaximum(self.max)
        widget.setAlignment(Qt.AlignRight)
        widget.setValue(self.value)
        return widget

    def get_value_from_widget(self, widget: QSpinBox) -> int:
        """
        Gets the integer value from the specified widget.
        :param widget: SpinBox to extract value from
        :return: Integer value
        """
        return widget.value()

    def is_valid(self, value: int) -> bool:
        """
        Verifies whether the value is valid for this parameter.
        :param value: Value to verify
        :return: True if value is valid, False otherwise
        """
        return self.min <= value <= self.max

    def export(self) -> str:
        """
        Exports the Verilog string representation of this parameter's value.
        :return: String representation
        """
        return f"{self.word_size}'d{self.value}"


class PercentParameter(IntParameter):
    """A parameter that can be set from 0 to 100%."""
    def __init__(self, *args, **kwargs):
        super(PercentParameter, self).__init__(v_min=0, v_max=100, word_size=8, *args, **kwargs)

    def copy(self) -> PercentParameter:
        """
        Creates a copy of itself.
        :return: Copy
        """
        return PercentParameter(name=self.name,
                                display_name=self.display_name,
                                description=self.description,
                                default=self.default,
                                value=self.value)

    def _base_widget(self) -> PercentSlider:
        """
        Creates a PercentSlider for this parameter.
        :return: PercentSlider widget
        """
        widget = PercentSlider(default=self.default)
        return widget

    def get_value_from_widget(self, widget: PercentSlider) -> int:
        """
        Extracts the value from the PercentSlider widget.
        :param widget: PercentSlider widget to extract value from
        :return: Extracted value
        """
        return widget.value


class StateParameter(IntParameter):
    """A parameter representing a raw FPGA device 24-bit state."""
    def __init__(self, *args, **kwargs):
        super(StateParameter,  self).__init__(v_min=0, v_max=0xFFFFFF, word_size=24, *args, **kwargs)

    def copy(self) -> StateParameter:
        """
        Creates a copy of itself.
        :return: Copy
        """
        return StateParameter(name=self.name,
                              display_name=self.display_name,
                              description=self.description,
                              default=self.default,
                              value=self.value)

    def _base_widget(self) -> StateSpinBox:
        """
        Creates a StateSpinBox for this parameter.
        :return: StateSpinBox widget
        """
        widget = StateSpinBox()
        widget.setMinimum(self.min)
        widget.setMaximum(self.max)
        widget.setAlignment(Qt.AlignRight)
        widget.setDisplayIntegerBase(16)
        widget.setPrefix("0x")
        widget.setValue(self.value)
        return widget

    def export(self) -> str:
        """
        Exports the parameter's value to its Verilog string representation.
        :return: Verilog string representation of value
        """
        return "24'h%06x" % self.value


class ColorParameter(StateParameter):
    """A parameter that represents a 24-bit hexadecimal color."""
    def __init__(self, *args, **kwargs):
        super(ColorParameter, self).__init__(*args, **kwargs)

    def copy(self) -> ColorParameter:
        """
        Creates a copy of itself.
        :return: Copy
        """
        return ColorParameter(name=self.name,
                              display_name=self.display_name,
                              description=self.description,
                              default=self.default,
                              value=self.value)

    def _base_widget(self) -> ColorButton:
        """
        Creates a ColorButton for this parameter.
        :return: ColorButton widget
        """
        col_tuple = (self.value >> 16, (self.value >> 8) % 256, self.value % 256)
        widget = ColorButton(default_color=col_tuple)
        return widget

    def get_value_from_widget(self, widget: ColorButton) -> Tuple[int, int, int]:
        """
        Extracts the value from the ColorButton widget.
        :param widget: ColorButton to extract value from
        :return: RGB color triplet
        """
        return (widget.color[0] << 16) + (widget.color[1] << 8) + widget.color[2]



from qtpy import QtWidgets, uic
from os import path


class BaseWindow(QtWidgets.QMainWindow):
    """Base class for all windows.

    Automatically assigns all of its children widgets as attributes to itself.
    """
    def __init__(self, ui_file: str, *args, **kwargs):
        super(BaseWindow, self).__init__(*args, **kwargs)
        uic.loadUi(path.join(path.dirname(__file__), "..", "res", "ui", ui_file), self)

        for child in self.findChildren(QtWidgets.QWidget):
            name = child.objectName()
            setattr(self, name, child)


class BaseDialog(QtWidgets.QDialog):
    """Base class for all dialogs.

    Automatically assigns all of its children widgets as attributes to itself.
    """
    def __init__(self, ui_file, *args, **kwargs):
        super(BaseDialog, self).__init__(*args, **kwargs)
        uic.loadUi(path.join(path.dirname(__file__), "..",  "res", "ui", ui_file), self)

        for child in self.findChildren(QtWidgets.QWidget):
            name = child.objectName()
            setattr(self, name, child)


class BaseWidget(QtWidgets.QWidget):
    """Base class for all widgets.

    Automatically assigns all of its children widgets as attributes to itself.
    """
    def __init__(self, ui_file, *args, **kwargs):
        super(BaseWidget, self).__init__(*args, **kwargs)
        uic.loadUi(path.join(path.dirname(__file__), "..",  "res", "ui", ui_file), self)

        for child in self.findChildren(QtWidgets.QWidget):
            name = child.objectName()
            setattr(self, name, child)



from qtpy.QtCore import Slot
from qtpy.QtWidgets import QSlider

from fpga_device_manager.windows.base import BaseWidget


class PercentSlider(BaseWidget):
    """A combination of a slider and a label for selecting a percentage value.

    The label automatically updates to display the slider's state.
    """
    def __init__(self, default: int = 0, v_min: int = 0, v_max: int = 100, *args, **kwargs):
        super(PercentSlider, self).__init__("label_slider.ui", *args, **kwargs)
        self.slider: QSlider
        self.value = default

        self.slider.setMinimum(v_min)
        self.slider.setMaximum(v_max)
        self.slider.setValue(default)
        self.refresh()

    def refresh(self) -> None:
        """Updates the label to reflect the currently chosen value."""
        self.label.setText(f"{self.value} %")

    @Slot(int)
    def on_slider_valueChanged(self, new_value: int):
        """Handler for choosing a new value on the slider.

        :param new_value: New value
        """
        self.value = new_value
        self.refresh()



from qtpy.QtWidgets import QSpinBox


class StateSpinBox(QSpinBox):
    """
    A QSpinBox that displays 24-bit hexadecimal values.
    """
    def __init__(self, *args, **kwargs):
        super(StateSpinBox, self).__init__(*args, **kwargs)
        self.setStyleSheet("QSpinBox {  font-family: monospace }")

    def textFromValue(self, v: int) -> str:
        """
        Converts a value into its text representation.
        :param v: Value to convert
        :return: Converted text
        """
        return ("%06x" % v).upper()



import os
import sys

from qtpy import QtWidgets

from fpga_device_manager import Devices, Inputs, Pins
from fpga_device_manager.windows.main_window import MainWindow

if __name__ == '__main__':
    base_data_path = os.path.join(os.path.dirname(__file__), "res", "data")

    Devices.init(os.path.join(base_data_path, "device_types.json"))
    Inputs.init(os.path.join(base_data_path, "input_types.json"))
    Pins.init(os.path.join(base_data_path, "pins.json"))

    # Create some directories if possible
    for dirname in ("configurations", "generated"):
        try:
            os.mkdir(dirname)
        except FileExistsError:
            pass

    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec_())



"""Module that manages FPGA pins."""
import json
from typing import Optional, Tuple, Iterable, Generator, Iterator

_pins = {}

PIN_RESERVED = 0
PIN_OCCUPIED_DEVICE = 1
PIN_OCCUPIED_INPUT = 2
PIN_AVAILABLE_PWM = 3
PIN_AVAILABLE = 4

_PIN_COLORS = {
    PIN_RESERVED: (80, 80, 80),
    PIN_OCCUPIED_DEVICE: (220, 80, 80),
    PIN_OCCUPIED_INPUT: (240, 150, 100),
    PIN_AVAILABLE_PWM: (180, 220, 180),
    PIN_AVAILABLE: (180, 180, 220)
}


class FPGAPin:
    """Represents a physical pin on the FPGA."""
    def __init__(self, name, display_name=None, supports_pwm=True, supports_output=True, supports_input=True,
                 fpga_addr=31, device_pin=None):
        self.name = name
        self.display_name = display_name or name
        self.supports_pwm = supports_pwm
        self.supports_output = supports_output
        self.supports_input = supports_input
        self.fpga_addr = fpga_addr
        self.device_pin = device_pin

    def is_assigned(self) -> bool:
        """
        Returns whether this pin is currently assigned to a device.
        :return: True if pin is assigned, False otherwise
        """
        return self.device_pin is not None

    def is_assignable(self, device_pin: Optional['DevicePin'] = None) -> bool:
        """
        Returns whether this pin is currently assignable. If a device pin is specified, it also checks
        for compatibility with it.
        :param device_pin: Optional device pin to check compatibility with
        :return: True if this pin is currently assignable, False otherwise
        """
        if device_pin is None:
            return not self.is_assigned()

        return not self.is_assigned() and device_pin.is_compatible_with(self)

    def is_input(self) -> bool:
        """
        Returns whether this pin is currently assigned to the pin of an Input device.
        :return: True if this pin is connected to an Input pin, False otherwise
        """
        from fpga_device_manager.Inputs import InputPin
        return self.is_assigned() and isinstance(self.device_pin, InputPin)

    def is_output(self) -> bool:
        """
        Returns whether this pin is currently assigned to the pin of an Output device.
        :return: True if this pin is connected to an Output pin, False otherwise
        """
        from fpga_device_manager.Devices import DevicePin
        return self.is_assigned() and isinstance(self.device_pin, DevicePin)

    def color(self) -> Tuple[int, int, int]:
        """
        Returns this pin's color for the preview image. The color depends on the current state of the pin
        and the device pin it might be connected to.
        :return: Pin's color, as a RGB color triplet (0..255)
        """
        return color(self.is_input() and PIN_OCCUPIED_INPUT or
                     self.is_output() and PIN_OCCUPIED_DEVICE or
                     self.supports_pwm and PIN_AVAILABLE_PWM or
                     PIN_AVAILABLE)


def init(filename: str) -> None:
    """
    Initializes the FPGA pin database by loading the pin configuration from the specified filename.
    The file must be in JSON format and contains pin settings, indexed by the FPGA's pin names.
    At the very least, each pin must contain the properties 'addr' (an integer describing the pin's internal adress)
    and 'pwm' (a Boolean value describing whether the pin supports PWM output).
    Optionally, it can also contain a 'name' attribute (a string overriding the FPGA pin name for the preview image)
    and 'output' and 'input' attributes (Boolean values describing whether the pin supports output or input,
    respectively, defaulting to True).

    Example of a configuration file:
    {
        "1": {
            "pwm": true,
            "addr": 1
        },
        "2": {
            "pwm": false,
            "input": false,
            "addr": 2
        }
        "99": {
            "pwm": false,
            "output": false,
            "name": "D1",
            "addr": 3
        }
    }

    :param filename: FPGA pin configuration file
    """
    global _pins
    _pins = {}

    try:
        with open(filename, "r") as file:
            pin_data = json.load(file)

        for pin, pin_args in pin_data.items():
            display_name = pin_args.get("name", pin)
            _pins[pin] = FPGAPin(name=pin,
                                 display_name=display_name,
                                 supports_pwm=pin_args["pwm"],
                                 supports_input=pin_args.get("input", True),
                                 supports_output=pin_args.get("output", True),
                                 fpga_addr=pin_args.get("addr", 31))

    except OSError as e:
        raise Exception("failed to load pin data file: %s" % e)


def get(pin_name: str) -> FPGAPin:
    """
    Looks up a FPGA pin by its name and returns it.
    If no pin with the specified name exists, this will throw a KeyError.
    :param pin_name: Name of the FPGA pin to look up
    :return: FPGA pin instance
    """
    return _pins[pin_name]


def get_next_available_pin(device_pin: 'DevicePin' = None) -> FPGAPin:
    """
    Searches for the next available pin that is compatible with the supplied device pin and returns it.
    If no pins are available, this raises an Exception.
    :param device_pin: Device pin to find assignable FPGA pin for
    :return: FPGA Pin instance that is available for assignment
    """
    for pin in _pins.values():
        if pin.is_assignable(device_pin):
            return pin
    raise Exception("No pins available")


def lookup(pin_display_name: str) -> FPGAPin:
    """
    Looks up a FPGA pin by its display name and returns it.
    The display name is the same as its name, unless it was overridden via configuration.
    If no pin with the specified display name exists, this will throw an Exception.
    :param pin_display_name: Display name of the FPGA pin to look up
    :return: FPGA pin instance
    """
    for pin in _pins.values():
        if pin.display_name == pin_display_name:
            return pin
    raise Exception("No such pin: %s" % pin_display_name)


def available(device_pin: 'DevicePin') -> Iterator[FPGAPin]:
    """
    A generator that supplies every FPGA pin that is assignable to the given device pin.
    :param device_pin: Iterator of assignable FPGA pins
    """
    for pin in _pins.values():
        if pin.is_assignable(device_pin):
            yield pin


def all() -> Iterable[FPGAPin]:
    """
    Returns an Iterable of all FPGA pins.
    :return: All FPGA pins
    """
    return _pins.values()


def assign(pin_name: str, device_pin: 'DevicePin') -> FPGAPin:
    """
    Attempts to assign the supplied device pin to the FPGA pin of the given name.
    This throws a KeyError if no FPGA pin with that name exists. Additionally, on incompatibility between the two pins
    an Exception is thrown.
    :param pin_name: Name of FPGA pin to assign the device pin to
    :param device_pin: Device pin to assign to the FPGA pin
    :return: Instance of FPGA pin
    """
    try:
        pin = _pins[pin_name]

        # Check if pins are already connected
        if pin.device_pin == device_pin:
            return pin

        # Check if pin is assignable to device pin
        if not pin.is_assignable(device_pin):
            raise Exception("Device pin %s is not assignable to pin %s" % (device_pin.name, pin_name))

        # Free previously assigned pin
        if device_pin.assigned_pin is not None:
            device_pin.assigned_pin.device_pin = None

        # Assign to new pin
        _pins[pin_name].device_pin = device_pin
        device_pin.assigned_pin = pin

        return pin

    except KeyError:
        raise Exception("Pin %s does not exist" % pin_name)


def clear(pin_name: str) -> FPGAPin:
    """
    Clears the FPGA pin with the given name of its associated pin. Does nothing if the FPGA pin is unassigned.
    :param pin_name: Name of FPGA pin to clear
    :return: FPGA Pin instance
    """
    try:
        pin = _pins[pin_name]
        pin.device_pin.assigned_pin = None
        pin.device_pin = None
        return pin
    except KeyError:
        raise Exception("Pin %s does not exist" % pin_name)


def color(pin_type: int) -> Tuple[int, int, int]:
    """
    Returns the color for the specified raw pin type.
    :param pin_type: Raw pin type ID
    :return: Color for pin type as RGB color triplet (0..255)
    """
    return _PIN_COLORS[pin_type]



from qtpy.QtCore import Slot
from qtpy.QtWidgets import QComboBox

from fpga_device_manager import Config
from fpga_device_manager.windows.base import BaseWidget


class DeviceAssociationWidget(BaseWidget):
    """A widget for choosing an output device to associate with."""
    def __init__(self, device_widget: 'InputWidget', *args, **kwargs):
        super(DeviceAssociationWidget, self).__init__("device_assoc.ui", *args, **kwargs)

        self.device_widget = device_widget
        self.device = device_widget.device
        self.main_window = device_widget.main_window

    def refresh(self) -> None:
        """Updates the widget. This displays all association possibilities."""
        self.combo_devices: QComboBox
        self.combo_devices.clear()

        if self.device.is_associated():
            self.combo_devices.addItem(self.device.associated_device.name, self.device.associated_device)
        else:
            self.combo_devices.addItem("", None)

        for supported_device in Config.outputs().filter(self.device.template.compatible_devices):
            if supported_device is not self.device.associated_device:
                self.combo_devices.addItem(supported_device.name, supported_device)

        if self.device.is_associated():
            self.lb_device.setStyleSheet("font-weight: normal; color: inherit")
        else:
            self.lb_device.setStyleSheet("font-weight: bold; color: red;")

    @Slot(int)
    def on_combo_devices_currentIndexChanged(self, index: int):
        """
        Handler for changing the device combo box.
        :param index: Index of chosen output device in combo box.
        """
        self.combo_devices: QComboBox
        dev_selected = self.combo_devices.itemData(index)
        if dev_selected is not None and dev_selected is not self.device.associated_device:
            self.device.associate(dev_selected)
            self.main_window.set_dirty()
            self.main_window.refresh()



from __future__ import annotations
from abc import ABC

from fpga_device_manager import Pins
from fpga_device_manager.Pins import FPGAPin
from fpga_device_manager.exceptions import PinNotAssignedError, UnknownPinException, DeviceLoadError


class DeviceBase(ABC):
    """Base class for all devices and device templates."""
    def __init__(self, dev_name, dev_type, uses_clk=False, pins=None, parameters=None):
        self.name = dev_name
        self.type = dev_type
        self.uses_clk = uses_clk
        self.pins = pins or {}
        self.parameters = parameters or {}


class DeviceBaseTemplate(DeviceBase, ABC):
    """Base class for device templates."""
    def __init__(self, dev_driver, description="", *args, **kwargs):
        super(DeviceBaseTemplate, self).__init__(*args, **kwargs)

        self.driver = dev_driver
        self.description = description


class DeviceBaseInstance(DeviceBase, ABC):
    """Base class for device instances."""
    def __init__(self, dev_id=None, *args, **kwargs):
        super(DeviceBaseInstance, self).__init__(*args, **kwargs)
        self.template = None
        self.dev_id = dev_id

    @staticmethod
    def create_from_template(dev_name: str, dev_type: int) -> DeviceBaseInstance:
        """
        Creates a new device instance from a template, identified by its index.
        :param dev_name: Name of new device
        :param dev_type: ID of template to base device off of
        :return: Newly created device
        """
        raise NotImplementedError

    @classmethod
    def load(cls, dev_data: dict) -> DeviceBaseInstance:
        """
        Loads a device from JSON data, loaded from a file.
        :param dev_data: JSON imported data structure
        :return: Loaded device object
        """
        try:
            obj = cls.create_from_template(dev_name=dev_data["name"],
                                           dev_type=dev_data["type"])
        except KeyError as e:
            raise DeviceLoadError(f'While setting up a device, a required key "{e}" was missing')

        for pin_name, pin_data in dev_data["pins"].items():
            if pin_name not in obj.pins:
                raise UnknownPinException(obj, pin_name)

            Pins.assign(pin_data["assigned_pin"], obj.pins[pin_name])
            obj.pins[pin_name].active_low = pin_data.get("active_low", False)

        for param_name, param_data in dev_data["parameters"].items():
            if param_name not in obj.parameters:
                continue

            obj.parameters[param_name].set_value(param_data)

        return obj

    def check_validity(self) -> bool:
        """
        Checks the device for proper configuration.
        Incorrect configurations throw an InvalidDeviceError.
        :return: True if device has been set up properly
        """
        raise NotImplementedError

    def save(self) -> dict:
        """
        Exports the device's properties to a JSON-serializable dict.
        :return: Device settings as dict
        """
        raise NotImplementedError


class DeviceBasePin(ABC):
    """Base class for device pins."""
    def __init__(self, name, device, active_low=False, assigned_pin=None):
        self.name = name
        self.device = device
        self.active_low = active_low
        self.assigned_pin = assigned_pin

    def copy(self, new_device: DeviceBaseInstance) -> DeviceBasePin:
        """
        Creates a copy of itself and assigns it to the specified device.
        :param new_device: Device to copy itself to
        :return: Copy of pin
        """
        raise NotImplementedError

    def is_assigned(self) -> bool:
        """
        Returns whether the pin is assigned to an FPGA pin.
        :return: True if pin is assigned, False otherwise
        """
        return self.assigned_pin is not None

    def is_compatible_with(self, fpga_pin: FPGAPin) -> bool:
        """
        Returns whether the pin is compatible with the specified FPGA pin.
        :param fpga_pin: Pin to check compatibility with
        :return: True if pins are compatible, False otherwise
        """
        raise NotImplementedError

    def check_validity(self) -> bool:
        """
        Checks whether the pin is configured properly.
        If not, throws a PinNotAssignedError.
        :return: True if pin is set up properly
        """
        if self.assigned_pin is None:
            raise PinNotAssignedError(self)

        return True



"""Module for managing Input device templates. Also contains classes for instances of Input devices."""
from __future__ import annotations
import json
from typing import Iterable

from fpga_device_manager.Devices import Output
from fpga_device_manager.device_base import DeviceBasePin, DeviceBaseTemplate, DeviceBaseInstance
from fpga_device_manager.device_params import DeviceParameter
from fpga_device_manager.exceptions import PinIncompatibilityError, InputNotAssociatedError, DeviceIncompatibilityError

_input_templates = {}


class InputPin(DeviceBasePin):
    """Base class for input pins."""
    def __init__(self, *args, **kwargs):
        super(InputPin, self).__init__(*args, **kwargs)

    def copy(self, new_device: DeviceBaseInstance) -> InputPin:
        """
        Creates a copy of itself and assigns it to a device.
        :param new_device:
        :return:
        """
        return InputPin(self.name,
                        device=self.device,
                        active_low=self.active_low,
                        assigned_pin=self.assigned_pin)

    def check_validity(self) -> bool:
        """
        Verifies the configuration of this pin. Will throw a PinIncompatibilityError if the assigned pin is somehow
        incompatible with this pin.
        :return: True if configuration is valid
        """
        super(InputPin, self).check_validity()

        if not self.is_compatible_with(self.assigned_pin):
            raise PinIncompatibilityError(self)

    def is_compatible_with(self, fpga_pin: 'FPGAPin') -> bool:
        """
        Returns whether this pin is compatible with the specified FPGA pin.
        :param fpga_pin: FPGA pin to check compatibility with
        :return: True if compatible, False otherwise
        """
        return fpga_pin.supports_output

    def __str__(self):
        return "Pin %s of input %s" % (self.name, self.device.name)


class InputTemplate(DeviceBaseTemplate):
    """Represents the blueprint on which Input devices can be based off of."""
    def __init__(self, compatible_devices=None, *args, **kwargs):
        super(InputTemplate, self).__init__(*args, **kwargs)

        self.compatible_devices = compatible_devices or []

    @staticmethod
    def load(dev_data: dict) -> InputTemplate:
        """
        Creates an InputTemplate based on loaded JSON data.
        :param dev_data: JSON dict template data
        :return: Template instance
        """
        obj = InputTemplate(dev_name=dev_data["name"],
                            dev_type=dev_data["type"],
                            dev_driver=dev_data["driver"],
                            uses_clk=dev_data["uses_clk"],
                            description=dev_data["description"],
                            compatible_devices=dev_data["compatible_devices"],
                            parameters={
                                param_name: DeviceParameter.load(param_name, param_data)
                                for param_name, param_data in dev_data.get("parameters", {}).items()
                            })

        for pin_name, pin_data in dev_data["pins"].items():
            obj.pins[pin_name] = InputPin(name=pin_name,
                                          device=obj,
                                          active_low=pin_data["active_low"])

        return obj


class Input(DeviceBaseInstance):
    """An input device or sensor."""
    def __init__(self, associated_device=None, *args, **kwargs):
        super(DeviceBaseInstance, self).__init__(*args, **kwargs)

        self.associated_device = associated_device

    @staticmethod
    def create_from_template(dev_name: str, dev_type: int) -> Input:
        """
        Creates an Input device instance based on an InputTemplate.
        :param dev_name: The name of the new device
        :param dev_type: The ID of the template to base the device off of
        :return: Input device instance
        """
        dev_template = get_template(dev_type)

        obj = Input(dev_name=dev_name,
                    dev_type=dev_template.type,
                    uses_clk=dev_template.uses_clk,
                    parameters=dev_template.parameters.copy())

        for name, pin in dev_template.pins.items():
            obj.pins[name] = pin.copy(obj)

        obj.template = dev_template

        return obj

    def save(self) -> dict:
        """
        Exports the input device's state into a format that is JSON serializable.
        :return: JSON serializable dict
        """
        return {
            "name": self.name,
            "type": self.type,
            "associated_device": self.associated_device.dev_id if self.is_associated() else None,
            "pins": {
                pin.name: {
                    "assigned_pin": pin.assigned_pin.name if pin.assigned_pin is not None else None,
                    "active_low": pin.active_low
                } for pin in self.pins.values()
            },
            "parameters": {name: param.value for name, param in self.parameters.items()}
        }

    def on_post_load(self, manager: 'DeviceManager', data: dict) -> None:
        """
        Hook that associates this Input device with an Output device after loading a configuration from file.
        :param manager: Device Manager instance
        :param data: Device data
        """
        self.associate(manager.devices[data["associated_device"]])

    def associate(self, device: Output) -> None:
        """
        Associates this Input device with an Output device.
        If the devices are incompatible with each other, this raises an
        :param device: Output device to associate with
        """
        # Remove old association
        self.deassociate()

        # Associate
        if device.type not in self.template.compatible_devices:
            raise DeviceIncompatibilityError(self, device)

        self.associated_device = device
        device.associated_inputs.append(self)

    def deassociate(self) -> None:
        """Removes the output device association of this device."""
        if self.associated_device is None:
            return

        self.associated_device.associated_inputs.remove(self)
        self.associated_device = None

    def is_associated(self) -> bool:
        """
        Returns whether this device is currently associated with an Output device.
        :return: True if device is associated, False otherwise
        """
        return self.associated_device is not None

    def check_validity(self) -> bool:
        """
        Verifies the validity of this device's configuration by checking whether it is associated with any Output device
        and whether all of its pins are valid
        :return: True if device is configured properly, False otherwise
        """
        if not self.is_associated():
            raise InputNotAssociatedError(self)

        for pin in self.pins.values():
            pin.check_validity()

        return True

    def __str__(self):
        return self.name


def init(filename: str) -> None:
    """
    Initializes the template module by loading and indexing all available Input device templates.
    :param filename: File to load template data from
    """
    global _input_templates
    try:
        with open(filename, "r") as file:
            data = json.load(file)
            for tpl_type, tpl_data in data.items():
                tpl_data["type"] = int(tpl_type)
                _input_templates[int(tpl_type)] = InputTemplate.load(tpl_data)

    except OSError as e:
        raise Exception("Couldn't open input template file: %s" % e)


def list_templates() -> Iterable[InputTemplate]:
    """
    Returns an Iterable of all Input device templates, indexed by ID.
    :return: Iterable of Input device templates
    """
    for tpl_id, tpl in _input_templates.items():
        yield tpl_id, tpl


def get_template(dev_type: int) -> InputTemplate:
    """
    Looks up an Input device template by ID.
    :param dev_type: ID of Input device template
    :return: Input device template instance
    """
    try:
        return _input_templates[dev_type]
    except KeyError:
        raise Exception("No such device type: %s" % dev_type)



from qtpy.QtCore import Slot, Signal
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QPushButton, QColorDialog
from typing import Tuple


class ColorButton(QPushButton):
    """A button for selecting a single color.

    The button's color will display the color chosen.  Additionally; the label reflects the hexcode of the color.
    """

    color_chosen = Signal(int, int, int)

    def __init__(self, default_color: Tuple[int, int, int], *args, **kwargs):
        super(ColorButton, self).__init__(*args, **kwargs)
        self.color = default_color
        self.clicked.connect(self.show_color_dialog)
        self.refresh()

    def refresh(self) -> None:
        """Updates the button's style to reflect the currently chosen color."""
        if self.color == (0, 0, 0):
            self.setText("(off)")
            self.setStyleSheet("")
        else:
            hex_col = ("#%02x%02x%02x" % self.color).upper()
            self.setText(hex_col)
            self.setStyleSheet(f"background-color: {hex_col};")

    @staticmethod
    def encode_qcolor(color: Tuple[int, int, int]) -> QColor:
        """
        Translates the given color triplet into a QColor.
        :param color: Color triplet (Tuple of R, G, B values ranging from 0 to 255 each)
        :return: QColor corresponding to given color triplet
        """
        return QColor(color[0], color[1], color[2])

    @staticmethod
    def decode_qcolor(color: QColor) -> Tuple[int, int, int]:
        """
        Translates the given QColor into a RGB color triplet.
        :param color: QColor to translate
        :return: RGB color triplet corresponding to given QColor
        """
        return color.red(), color.green(), color.blue()

    @Slot()
    def show_color_dialog(self):
        """Displays a dialog for choosing a new color."""
        color_dlg = QColorDialog()
        color_dlg.setCurrentColor(self.encode_qcolor(self.color))
        if color_dlg.exec() == color_dlg.Accepted:
            self.color = self.decode_qcolor(color_dlg.currentColor())
            print(self.color)
            self.refresh()



from typing import Dict, Optional, Iterable

from fpga_device_manager import Pins, Constants
from fpga_device_manager.device_base import DeviceBaseInstance


class DeviceManager:
    """Manages a type of device."""
    def __init__(self, device_class, max_devices: Optional[int] = None):
        self.devices = {}  # type: Dict[int, DeviceBaseInstance]
        self.max_devices = max_devices or Constants.MAX_DEVICES
        self.available_ids = list(range(0, self.max_devices))

        self.device_class = device_class

    def has_devices(self):
        """
        Returns whether this manager has at least one device.
        :return: True if there are devices, False otherwise
        """
        return len(self.devices) > 0

    def has_max_devices(self):
        """
        Returns whether this manager has no more room for devices.
        :return: True if number of devices is maxed out, False otherwise
        """
        return len(self.devices) == self.max_devices

    def add_device(self, device: DeviceBaseInstance, force_id: Optional[int] = None) -> None:
        """
        Adds a new device.
        :param device: The device to add
        :param force_id: Forces a device ID, if specified. Otherwise just uses the next free one. If the requested ID
                         is not available, this throws an Exception.
        """
        if force_id is not None:
            try:
                self.available_ids.remove(force_id)
                device_id = force_id
            except ValueError:
                raise Exception("The requested device ID %d is not available" % force_id)
        else:
            device_id = self.available_ids.pop(0)

        device.dev_id = device_id
        self.devices[device_id] = device

    def move_device(self, old_id: int, new_id: int) -> None:
        """
        Attempts to move an existing device from one ID to another.
        If the device with the specified ID doesn't exist or the target ID is already occupied, this throws an
        Exception.
        :param old_id: ID of the device to move
        :param new_id: New ID to assign device to
        """
        if new_id not in self.available_ids:
            raise Exception("ID %d is not available" % new_id)

        try:
            self.devices[new_id] = self.devices.pop(old_id)
            self.devices[new_id].dev_id = new_id
            self.available_ids.remove(new_id)
            self.available_ids.append(old_id)
            self.available_ids.sort()
        except KeyError:
            raise Exception("ID %d is not used" % old_id)

    def remove_device(self, device_id: int) -> None:
        """
        Removes a device with the specified ID. If no device with that ID exists, this throws an Exception.
        :param device_id: ID of device to remove
        """
        try:
            # Get device
            device = self.devices[device_id]

            # Return device ID
            self.available_ids.append(device_id)
            self.available_ids.sort()

            # Clear pins
            for pin in device.pins.values():
                if pin.is_assigned():
                    Pins.clear(pin.assigned_pin.name)

            # Remove device
            del self.devices[device_id]

        except KeyError:
            raise Exception("No device with ID %d" % device_id)

    def clear(self) -> None:
        """Removes all devices from this manager, effectively resetting it."""
        for dev_id in self.devices.copy().keys():
            self.remove_device(dev_id)

        self.devices.clear()
        self.available_ids = list(range(0, self.max_devices))

    def get(self, device_id: int) -> DeviceBaseInstance:
        """
        Looks up a device based on its ID and returns it.
        If no device with that ID exists, throws an Exception.
        :param device_id: ID of device to get
        :return: Device
        """
        try:
            return self.devices[device_id]
        except KeyError:
            raise Exception("no device with ID %d exists" % device_id)

    def all_devices(self) -> Iterable[DeviceBaseInstance]:
        """
        Returns all devices managed by this manager.
        :return: All devices
        """
        return self.devices.values()

    def filter(self, dev_types: Iterable[int]) -> Iterable[DeviceBaseInstance]:
        """
        Returns all devices that are of one of the types specified.
        :param dev_types: Filtered list of devices
        """
        for dev in self.devices.values():
            if dev.type in dev_types:
                yield dev

    def load(self, dev_list: dict) -> None:
        """
        Loads devices from JSON data.
        :param dev_list: JSON-loaded dict data
        """
        for dev_id, dev_data in dev_list.items():
            device = self.device_class.load(dev_data)
            self.add_device(device, force_id=int(dev_id))

    def save(self) -> Dict[int, Dict]:
        """
        Exports all devices into a JSON-serializable dictionary.
        :return: Dict to be serialized
        """
        return {dev_id: device.save() for dev_id, device in self.devices.items()}



import setuptools

from fpga_device_manager.Constants import APP_VERSION

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='fpga-device-manager',
    version=APP_VERSION,
    packages=setuptools.find_packages(),
    url='',
    license='MIT',
    author='Hannes Preiss',
    author_email='sophie@sophieware.net',
    description='A GUI for managing FPGA Device configurations.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=['moody-templates', 'PySide2', 'QtPy'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ]
)


from qtpy.QtCore import Slot
from qtpy.QtWidgets import QComboBox

from fpga_device_manager import Pins
from fpga_device_manager.device_base import DeviceBasePin
from fpga_device_manager.exceptions import DeviceInvalidError
from fpga_device_manager.windows.base import BaseWidget
from fpga_device_manager.Devices import DevicePin


class DevicePinWidget(BaseWidget):
    """A widget representing a device pin."""
    def __init__(self, device_widget: 'DeviceBaseWidget', device_pin: DeviceBasePin, *args, **kwargs):
        super(DevicePinWidget, self).__init__("device_pin.ui", *args, **kwargs)

        self.device_widget = device_widget
        self.main_window = device_widget.main_window
        self.device_pin = device_pin

        self.lb_pin_name.setText(self.device_pin.name)
        self.chk_active_low.setChecked(self.device_pin.active_low)

        if isinstance(self.device_pin, DevicePin):
            self.chk_pwm.setChecked(self.device_pin.requires_pwm)

        self.refresh()

    def refresh(self) -> None:
        """Updates the widget to reflect the current state of the pin."""
        self.combo_pins: QComboBox
        self.combo_pins.clear()

        if self.device_pin.assigned_pin is not None:
            self.combo_pins.addItem(self.device_pin.assigned_pin.display_name)
        else:
            self.combo_pins.addItem("")

        self.combo_pins.addItems(pin.display_name for pin in Pins.available(self.device_pin))

        try:
            self.device_pin.check_validity()
            self.lb_pin_name.setStyleSheet("")
        except DeviceInvalidError:
            self.lb_pin_name.setStyleSheet("font-weight: bold; color: red")

    @Slot(int)
    def on_chk_active_low_stateChanged(self, state: int):
        """
        Handler for changing the state of the active low checkbox.
        :param state: The checkbox's new state.
        """
        self.main_window.set_dirty()
        self.device_pin.active_low = state == 2

    @Slot(str)
    def on_combo_pins_currentTextChanged(self, new_pin: str):
        """
        Handler for changing the pin assignment combo box.
        :param new_pin: The selected FPGA pin to assign to.
        """
        if new_pin != '':
            if self.device_pin.assigned_pin is None or new_pin != self.device_pin.assigned_pin.display_name:
                pin_name = Pins.lookup(new_pin).name
                Pins.assign(pin_name, self.device_pin)
                self.main_window.set_dirty()
                self.main_window.refresh()



"""Application constants."""

"""The application name. This is displayed on the window's title and in every exported file."""
APP_NAME = "FPGA Device Manager"

"""Application version. Displayed next to the name."""
APP_VERSION = "0.2.4"

"""Default setting for maximum number of devices."""
MAX_DEVICES = 255



from typing import Union

from qtpy.QtCore import Slot
from qtpy.QtWidgets import QComboBox, QVBoxLayout

from fpga_device_manager import Config
from fpga_device_manager.device_base import DeviceBaseInstance
from fpga_device_manager.device_manager import DeviceManager
from fpga_device_manager.exceptions import DeviceInvalidError
from fpga_device_manager.widgets.device_assoc import DeviceAssociationWidget
from fpga_device_manager.widgets.device_pin import DevicePinWidget
from fpga_device_manager.windows.base import BaseWidget
from fpga_device_manager.windows.device_settings import DeviceSettingsWindow


class DeviceBaseWidget(BaseWidget):
    """A widget representing a device and its state, including several children widgets."""
    def __init__(self, device: Union['Device', 'Output'], main_window: 'MainWindow', manager: DeviceManager,
                 *args, **kwargs):
        super(DeviceBaseWidget, self).__init__("device.ui", *args, **kwargs)

        self.device = device
        self.main_window = main_window
        self.manager = manager

        self.setup_pin_widgets()
        self.refresh()

    def setup_pin_widgets(self) -> None:
        """Sets up widgets for the device's pins."""
        for pin in self.device.pins.values():
            self.lyo_pins.addWidget(DevicePinWidget(device_widget=self, device_pin=pin))

    def refresh(self) -> None:
        """Updates the widget and its children to reflect its current state."""
        self.group.setTitle(self.device.name)

        self.combo_ids: QComboBox
        self.combo_ids.clear()
        self.combo_ids.addItem(str(self.device.dev_id))
        self.combo_ids.addItems(str(i) for i in Config.outputs().available_ids)

        try:
            self.device.check_validity()
            self.group.setStyleSheet("")
        except DeviceInvalidError:
            self.group.setStyleSheet("QGroupBox::title {font-weight: bold; color: red;}")

        self.lyo_pins: QVBoxLayout

        # Refresh children
        for i in range(self.lyo_pins.count()):
            self.lyo_pins.itemAt(i).widget().refresh()

    @Slot()
    def on_btn_rename_clicked(self):
        """Handler for clicking the Rename button."""
        window = DeviceSettingsWindow(self.device)
        window.exec()
        self.refresh()

    @Slot()
    def on_btn_delete_clicked(self):
        """Handler for clicking the Delete button."""
        self.manager.remove_device(self.device.dev_id)
        self.main_window.remove_widget(self.device)

    @Slot(str)
    def on_combo_ids_currentTextChanged(self, new_id: str):
        """
        Handler for changing the device ID combo box.
        :param new_id: Selected device ID
        """
        if new_id != '' and int(new_id) != self.device.dev_id:
            self.manager.move_device(self.device.dev_id, int(new_id))
            self.main_window.refresh()


class OutputWidget(DeviceBaseWidget):
    """A widget representing an output device."""
    def __init__(self, *args, **kwargs):
        super(OutputWidget, self).__init__(manager=Config.outputs(), *args, **kwargs)

    @Slot()
    def on_btn_delete_clicked(self):
        """Handler for clicking the Delete button."""
        self.device.remove_associations()
        super(OutputWidget, self).on_btn_delete_clicked()


class InputWidget(DeviceBaseWidget):
    """A widget representing an input device."""
    def __init__(self, *args, **kwargs):
        super(InputWidget, self).__init__(manager=Config.inputs(), *args, **kwargs)

    def setup_pin_widgets(self) -> None:
        """Sets up widgets for the device's pins and output association."""
        self.lyo_pins.addWidget(DeviceAssociationWidget(device_widget=self))
        super(InputWidget, self).setup_pin_widgets()



"""
This module maintains a device configuration. It provides a device manager for output and input modules each.
"""
import json
import os
from datetime import datetime
from typing import Optional

import moody

from fpga_device_manager import Pins, Constants
from fpga_device_manager.Devices import Output
from fpga_device_manager.Inputs import Input
from fpga_device_manager.device_manager import DeviceManager
from fpga_device_manager.exceptions import DeviceInvalidError, InvalidConfigError

_output_mgr = DeviceManager(device_class=Output)
_input_mgr = DeviceManager(device_class=Input)


def clear() -> None:
    """Resets the configuration by clearing both device managers."""
    _output_mgr.clear()
    _input_mgr.clear()


def outputs() -> DeviceManager:
    """
    Returns the device manager for output devices.
    :return: Output device manager
    """
    return _output_mgr


def inputs() -> DeviceManager:
    """
    Returns the device manager for input devices.
    :return: Input device manager
    """
    return _input_mgr


def load(cfg_filename: str, max_devices: Optional[int] = None) -> None:
    """
    Loads a device configuration from a specified file.
    :param cfg_filename: Name of the file to load the configuration from.
    :param max_devices: Optionally specifies an upper limit for devices.
    """
    global _output_mgr, _input_mgr

    max_devices = max_devices or Constants.MAX_DEVICES

    _output_mgr = DeviceManager(device_class=Output, max_devices=max_devices)
    _input_mgr = DeviceManager(device_class=Input, max_devices=max_devices)

    with open(cfg_filename, "r") as file:
        device_file_data = json.load(file)

    for mgr, dev_list in ((_output_mgr, device_file_data["devices"]), (_input_mgr, device_file_data["inputs"])):
        mgr.load(dev_list)

        # Re-associate inputs
        for dev_input in _input_mgr.all_devices():
            associated_device_id = device_file_data["inputs"][str(dev_input.dev_id)]["associated_device"]
            if associated_device_id is not None:
                device = _output_mgr.get(associated_device_id)
                dev_input.associate(device)


def save(cfg_filename: str) -> None:
    """
    Saves the current device configuration to a file.
    :param cfg_filename: Name of file to save the configuration to.
    """
    data = {
        "devices": _output_mgr.save(),
        "inputs": _input_mgr.save()
    }

    with open(cfg_filename, "w") as file:
        json.dump(data, file, indent=2)


def check() -> None:
    """
    Performs sanity checks on the current configuration. In particular, this checks all device pins for proper
    assignment and all input/output device associations.

    If all checks pass, returns nothing. Otherwise, an InvalidConfigError is returned, passing all error messages along
    with it.
    """
    errors = []

    if len(list(_output_mgr.all_devices())) < 1:
        errors.append("A configuration needs at least one appliance")

    if len(list(_input_mgr.all_devices())) < 1:
        errors.append("A configuration needs at least one sensor")

    for mgr in (_output_mgr, _input_mgr):
        for device in mgr.all_devices():
            try:
                device.check_validity()
            except DeviceInvalidError as e:
                errors.append(str(e))

    if len(errors) > 0:
        raise InvalidConfigError(errors)


def export(out_path: str) -> None:
    """
    Exports the current device configuration to Verilog code.
    :param out_path: Path to write Verilog files to
    """
    # Gather devices
    output_devices = list(_output_mgr.all_devices())
    input_devices = list(_input_mgr.all_devices())
    output_max_id = max([dev.dev_id for dev in output_devices])
    input_max_id = max([dev.dev_id for dev in input_devices])

    # Gather device types
    output_device_types = {dev.dev_id: dev.type for dev in output_devices}
    input_device_types = {dev.dev_id: dev.type for dev in input_devices}

    # Gather input device mappings
    input_assignments = {dev.dev_id: dev.associated_device.dev_id for dev in input_devices}

    # Gather and enumerate used pins
    output_pins = [pin for pin in Pins.all() if pin.is_output()]
    input_pins = [pin for pin in Pins.all() if pin.is_input()]

    # Reverse lookup tables for pin indices
    output_pin_indices = {pin.name: index for index, pin in enumerate(output_pins)}
    input_pin_indices = {pin.name: index for index, pin in enumerate(input_pins)}

    # Gather unused pins
    unused_pins = [pin for pin in Pins.all() if not pin.is_assigned()]

    tpl_loader = moody.make_loader(os.path.join(os.path.dirname(__file__), "res", "vtemplates"))

    out_generator = f"{Constants.APP_NAME} {Constants.APP_VERSION}"
    out_timestamp = datetime.strftime(datetime.now(), "%Y/%m/%d %H:%M:%S")
    exports = {
        "devices_out.v": tpl_loader.render("devices_out.v.tpl",
                                           generator=out_generator,
                                           timestamp=out_timestamp,
                                           devices=output_devices,
                                           pin_ids=output_pin_indices,
                                           unused_pins=unused_pins),
        "devices_in.v": tpl_loader.render("devices_in.v.tpl",
                                          generator=out_generator,
                                          timestamp=out_timestamp,
                                          devices=input_devices,
                                          pin_ids=input_pin_indices,
                                          unused_pins=unused_pins),
        "types.v": tpl_loader.render("types.v.tpl",
                                     generator=out_generator,
                                     timestamp=out_timestamp,
                                     max_output_id=output_max_id,
                                     device_types=output_device_types,
                                     input_types=input_device_types,
                                     max_input_id=input_max_id,
                                     input_assignments=input_assignments),
        "pins_out.v": tpl_loader.render("pins_out.v.tpl",
                                        generator=out_generator,
                                        timestamp=out_timestamp,
                                        device_count=output_max_id + 1,
                                        pins=output_pins),
        "pins_in.v": tpl_loader.render("pins_in.v.tpl",
                                       generator=out_generator,
                                       timestamp=out_timestamp,
                                       device_count=len(input_devices),
                                       pins=input_pins),
        "smarthome.ldf": tpl_loader.render("smarthome.ldf.tpl",
                                           device_types=output_device_types.values(),
                                           input_types=input_device_types.values())
    }

    if not os.path.exists(out_path):
        os.mkdir(out_path)

    for filename, text in exports.items():
        filepath = os.path.join(out_path, filename)
        print(f"Writing to {filepath}...")
        with open(filepath, "w") as file:
            file.write(text)



"""Contains utility functions for creating popup messages."""
from typing import Callable

from qtpy.QtWidgets import QMessageBox
import traceback


def _prepare_popup(title: str, message: str, additional_text: str, icon, buttons) -> QMessageBox:
    popup = QMessageBox()
    popup.setWindowTitle(title)
    popup.setText(message)
    popup.setInformativeText(additional_text)
    popup.setIcon(icon)
    popup.setStandardButtons(buttons)
    return popup


def info(title: str, message: str, additional_text: str = "") -> None:
    """
    Creates and displays a notification popup.
    :param title: Title of popup window
    :param message: Popup message text
    :param additional_text: Second popup message text
    """
    _prepare_popup(title=title,
                   message=message,
                   additional_text=additional_text,
                   icon=QMessageBox.Information,
                   buttons=QMessageBox.Ok).exec_()


def alert(title: str, message: str, additional_text: str = "") -> None:
    """
    Creates and displays an alert popup.
    :param title: Title of popup window
    :param message: Popup message text
    :param additional_text: Second popup message text
    """
    _prepare_popup(title=title,
                   message=message,
                   additional_text=additional_text,
                   icon=QMessageBox.Warning,
                   buttons=QMessageBox.Ok).exec_()


def confirm(title: str,
            message: str,
            additional_text: str = "",
            on_yes: Callable[[], None] = lambda: None,
            on_no: Callable[[], None] = lambda: None) -> None:
    """
    Creates and displays a confirmation popup. Calls a different callable depending on whether the user accepts or not.
    :param title: Title of popup window
    :param message: Popup message text
    :param additional_text: Second popup message text
    :param on_yes: Callable that gets called when the user accepts the confirmation
    :param on_no: Callable that gets called when the user rejects or closes the confirmation
    """
    popup = _prepare_popup(title=title,
                           message=message,
                           additional_text=additional_text,
                           icon=QMessageBox.Question,
                           buttons=QMessageBox.Yes | QMessageBox.No)

    on_yes() if popup.exec_() == QMessageBox.Yes else on_no()


def error(title: str, ex: Exception) -> None:
    """
    Creates and displays an error popup that shows the contents of an exception.
    Afterwards, the exception is raised once more.
    :param title: Title of popup window
    :param ex: Exception to display
    """
    alert(title=title,
          message=str(ex),
          additional_text=traceback.format_exc())
    raise
